;TI-89/92+ Hardware 2.00 AMS 2.0x TSR support v.1.03 HEADER FILE VERSION
;Copyright (C) Kevin Kofler, 2001
;This is the source for the C version of the include file.
;To use it, use #include "h220xtsr.h" and call the function h220xTSR();.
;Use it only in your main function and if you do not use it at the end, you will
;have to relocate your program to the ghost space if it was not already there.
;For an event hook installer, I recommend the following sequence:
;void _main(void)
;{
;...
;if (!h220xTSR()) {ST_helpMsg("Memory error message");return;};
;*EV_hook=...;
;ST_helpMsg("Success message");
;}
;please read the readme file before using this file
;Please DO NOT distribute modified versions without my permission! Just
;include it as is and it should work.

 include "OS.h" ;You need the version included with TeOS
                ;or the one given with TIGCC 0.91 or
                ;higher (because of the ROM_CALL2 macro).
                ;If you want to use the TIGCC IDE under
                ;TIGCC 0.9 or lower, you have to copy
                ;this file to the TIGCC\INCLUDE\ASM
                ;folder (replacing the older version
                ;given with old versions of TIGCC).
 xdef _nostub ;no kernel required
 xdef _ti89 ;This is for the TI-89.
 xdef _ti92plus ;This is also for the TI-92+.
 xdef _main ;This is a program, not a library.

h220xTSR_FAILED equ 0
h220xTSR_SUCCESS equ 1
h220xTSR_SUCCESS_UPDATE equ 2
h220xTSR_HW1 equ -1
h220xTSR_AMS1 equ -2
h220xTSR_PATCHED equ -3
h220xTSR_ALREADY equ -4

_main:
 movem.l a2-a6/d3-d7,-(a7) ;save all registers
;detect the hardware version
;Thanks to Julien Muchembled for the C version listed in the TI-GCC FAQ.
;Converted to assembly by Kevin Kofler
 move.l $c8,d0
 and.l #$600000,d0 ;get the ROM base
 move.l d0,a0
 move.l 260(a0),a1 ;get the pointer to the hardware parameter block
 add.l #$10000,a0
 cmp.l a0,a1 ;check if the hardware parameter block is near enough
 bcc hw1 ;if it is too far, it is HW1
 cmp.w #22,(a1) ;check if the parameter block contains the hardware version
 bls hw1 ;if it is too small, it is HW1
 cmp.l #2,22(a1) ;check the hardware version
 bne hw1 ;if not 2, it is HW1 (or an unknown hardware version)
;end of hardware version detection routine
 move.l $c8,a0
 cmp.l #1000,-4(a0) ;check if more than 1000 entries
 bcs ams1 ;if no, it is AMS 1.xx
 ROM_CALL2 EX_stoBCD
 tst.w 92(a4) ;check for HW2Patch (ROM resident)
 beq patched ;if it is installed, refuse installation
 cmp.l #$100,$ac ;check for HW2Patch (RAM resident)
 beq patched ;if it is installed, refuse installation
 moveq.l #0,d4 ;set to "first install" mode
 cmp.l #$200000,$ac ;check if trap #$b points to the ROM
 bcs already ;if no, it is already installed
 cmp.l #$200000,$90 ;check if trap #4 points to the ROM
 bcs trap4_already ;if no, it is already hooked
 bsr enter_the_ghost_space ;unprotect and enter the ghost space
continue:
 lea.l oldtrapb(PC),a0
 move.l $ac,(a0) ;save the old trap #$B and call it from the new one
 lea.l oldtrap4(PC),a0
 move.l $90,(a0) ;save the old trap #4 and call it from the new one
 move.l #(endtrap4-newtrapb),-(a7)
 ROM_CALL HeapAllocHigh ;allocate a handle for the new trap #$B
 addq.l #2,a7
 tst.w d0 ;check if NULL handle
 beq nomem ;if yes, display "no mem" error message
 move.w d0,(a7)
 ROM_CALL HeapDeref ;get ("dereference") the address
 move.l a0,a3 ;save the address to a3 for further use
 subq.l #2,a7
 move.l #(endtrap4-newtrapb),(a7)
 pea.l newtrapb(PC)
 move.l a0,-(a7)
 ROM_CALL memcpy ;copy the trap #$B routine to the handle
 lea.l 12(a7),a7 ;shorter as add(a).l #12,a7
 add.l #$40008,a3 ;add: - $40000 for HW2 AMS 2.0x without
                  ;       HW2Patch
                  ;     - $8 to skip the uninstall
                  ;       information (need the start of
                  ;       the executable code here)
 move.l a3,$400ac ;install the trap #$B
 adda.w #newtrap4-newtrapb,a3 ;new trap #$4
 move.l a3,$40090 ;install the trap #$4
 tst.b d4 ;check if update mode
 bne update
 moveq.l #h220xTSR_SUCCESS,d0
return:
 movem.l (a7)+,a2-a6/d3-d7 ;restore all registers
 move.l (a7)+,a0           ; pick up the return address and make
 add.l #$40000,a0          ;   it "ghost" (very important)
 jmp (a0)                  ; return, but now in the ghost space
update:
 moveq.l #h220xTSR_SUCCESS_UPDATE,d0
 bra return
hw1:
 moveq.l #h220xTSR_HW1,d0
 bra return
ams1:
 moveq.l #h220xTSR_AMS1,d0
 bra return
patched:
 moveq.l #h220xTSR_PATCHED,d0
 bra return
already:
 move.l $ac,a0 ;check for previous h220xTSR installation
 cmp.l #$40000,a0 ;if <$40000, it is not h220xTSR (because that would not work
 bcs already2     ;reliably on HW2)
 sub.l #$40008,a0  ;remove: * $40000 for HW2 AMS 2.0x without HW2Patch
                   ;        * $8 to get the beginning of the handle (if
                   ;          compatible)
 cmp.l #'2TSR',(a0) ;check signature
 bne already2 ;if it is not the old one, refuse installation
 moveq.l #1,d4 ;set to update mode
 cmp.l #$200000,$90 ;check if trap #4 points to the ROM
 bcs trap4_already ;if no, it is already hooked
 move.l 4(a0),$400ac ;restore old trap #$b
 move.l a0,-(a7)
 ROM_CALL HeapPtrToHandle ;get the handle number
 addq.l #2,a7
 move.w d0,(a7)
 ROM_CALL HeapFree ;free (unallocate) the handle
 addq.l #2,a7
 bra continue ;continue the installation
already2:
 moveq.l #h220xTSR_ALREADY,d0
 bra return
trap4_already:
 move.l #$10000,-(a7)
 pea.l trap4DL(PC)
 pea.l trap4DLT(PC) ;"trap 4 already hooked" error message 
 bra showmsg
nomem:
 subq.l #2,a7 ;adjust the stack
 move.l #$10000,(a7)
 pea.l nomemDL(PC)
 pea.l nomemDLT(PC) ;"memory" error message 
showmsg:
 ROM_CALL DlgMessage ;display error dialog
 lea.l 12(a7),a7 ;restore the stack
 moveq.l #h220xTSR_FAILED,d0
 bra return

;Thanks to Zeljko Juric for this one! (slightly adapted to fit my needs)
enter_the_ghost_space:
 lea -20(a7),a7              ; make some space on stack
 move.l #$3E000,a3          ; prepare area at $3E000 for unprotecting
 move.w (a3),-(a7)          ; save four bytes at $3E000, because small
 move.l 2(a3),-(a7)         ;   link code will be formed there
 move.l $C8,d0              ; TIOS jump table in d0
 andi.l #$600000,d0         ; =$200000 for TI89, =$400000 for TI92+
 addi.l #$20000,d0          ; now d0 points somewhere in TIOS
 move.l d0,12(a7)           ; fool the routine to believe that it is
 move.l d0,16(a7)           ;   called somewhere from TIOS
 trap #$C                   ; this trap only enters supervisor mode
 move.w #$2700,sr           ; disable ints
 move.l #$F,d3              ; prepare for calling trap #$B indirectly
 move.l a3,-(a7)            ; push $3E000 on the stack (new PC)
 clr.w  -(a7)               ; push 0 on the stack (new SR)
 lea nx(pc),a0              ; put the address of label "nx" in a0
 adda.l #$40000,a0          ; add $40000 to make it "ghost"
 move.w #$4EF9,(a3)         ; this and the following instruction
 move.l a0,2(a3)            ;   create "jmp nx+40000" at $3E000
 move.l $AC,a0              ; the vector for trap #$B
 jmp (a0)                   ; enter into trap #$B indirectly

; Now, trap #$B will unprotect the 8K/24K area starting from $3E000.
; This includes the last page of the RAM too, so executing from the
; the ghost space is enabled too. The stack is set up on such way,
; that when the trap #$B is finished, the "rte" will transfer 
; execution to the address $3E000. But, there you will find 
; jmp nx+40000, so the execution continues at the label "nx", now
; in the "ghost space":

nx:
 move.l (a7)+,$3E002        ; restore the original content at
 move.w (a7)+,$3E000        ;   address $3E000
 lea 20(a7),a7              ; free the stack space
 move.l (a7)+,a0            ; pick up the return address and make
 adda.l #$40000,a0          ;   it "ghost" (very important)
 jmp (a0)                   ; return, but now in the ghost space

newtrapb:
 dc.b '2Tsr' ;signature
oldtrapb: dc.l 0 ;This is a placeholder for the original trap #$B address.
 cmp.l #$f,d3 ;check if function $f
 bne callold ;if no, just call the original trap #$B
 cmp.l #$200000,2(a7) ;check if enter_ghost_space
 bcs alreadyghost ;if yes, do nothing (program execution will be at PC+$80000)
                  ;which is still in the ghost space (ghost of ghost)
 cmp.l a2,a0 ;check if Exec command
 bne exec ;if yes, special treatement
;The address of the program to execute is stored in a2 by AMS 2.0x.
 cmp.l #$40000,a2 ;check if already in the ghost space
 bcc alreadyghost ;if yes, do nothing
 add.l #$40000,a2 ;add $40000 to a2 if it is not already in the ghost space
 moveq.l #0,d0 ;get the end of the program
 move.w -2(a2),d0
 add.l a2,d0
 subq.l #1,d0
 move.l d0,-(a7)
 move.l a2,-(a7)
common: ;common for ASM and Exec
 move.l $C8,a0
 move.l EX_patch*4(a0),a0 ;relocate the program again:
 jsr (a0)                 ;call EX_patch without destroying a4
 addq.l #8,a7
alreadyghost:
 rte
exec:
 cmp.l a5,a0 ;check if really Exec command
 bne alreadyghost ;if no, do nothing!!!
;The address of the STR_TAG of the exec string is stored in a2 by AMS 2.0x.
;The program address is stored in a5 by AMS 2.0x.
 add.l #$40000,a5 ;add $40000 to a5 if it is not already in the ghost space
 moveq.l #0,d0 ;get the end of the program
 lea.l -1(a2),a1 ;not to change a2 in order to respect the calling convention
execnext:
 addq.l #1,d0
 tst.b -(a1)
 bne execnext
 lsr.l #1,d0
 add.l a5,d0
 move.l d0,-(a7)
 move.l a5,-(a7) 
 bra common
callold:
 move.l oldtrapb(PC),-(a7) ;get the original trap #$B address (on the stack)
 rts ;jump to it (from the stack, in order not to destroy any registers)
newtrap4:
 dc.b '2Tsr' ;signature
oldtrap4: dc.l 0 ;This is a placeholder for the original trap #$4 address.
 move.l usp,a0
 move.l 2(a7),-(a0) ;push return address on user stack
 move.w (a7),-(a0) ;save return SR on user stack
 clr.w (a7) ;set return SR to "user mode"
 move.l a3,-(a0) ;save a3
 move.l d3,-(a0) ;save d3
 move.l $ac,-(a0) ;save the new trap #$B
 lea.l reenter_ghost_space-finishtrap4(a0),a0 ;save unprotected zone to user stack
 move.l #finishtrap4-reenter_ghost_space,-(a7)
 pea.l $5fc0
 move.l a0,-(a7)
 move.l $c8,a0
 move.l memcpy*4(a0),a0 
 jsr (a0) ;call memcpy without destroying a4
 move.l (a7)+,a0 ;restore a0
 addq.l #4,a7 ;restore a7
 lea.l finishtrap4(PC),a1 ;push end of routine to the user stack
 move.l a1,-(a0)
 move.l oldtrapb(PC),$400ac ;reset it to the original value
 move.l a0,usp
 pea.l reenter_ghost_space(PC)  ;copy middle of routine to
 pea.l $5fc0                    ;unprotected zone
 move.l $c8,a0
 move.l memcpy*4(a0),a0 
 jsr (a0)                 ;call memcpy without destroying a4
 lea.l 12(a7),a7
 move.l #$5fc0,2(a7) ;change return address to middle of routine
 move.l oldtrap4(PC),-(a7) ;get the original trap #$4 address (on the stack)
 rts ;jump to it (from the stack, in order not to destroy any registers)
reenter_ghost_space: ;This will be in user mode! It will be at $5fc0!
                     ;This is again an adaptation of Zeljko Juric's routine.
 move.l (a7),a0 ;a0=return address
 lea -16(a7),a7 ;make some space on stack
 move.l $C8,d0 ;TIOS jump table in d0
 andi.l #$600000,d0 ;=$200000 for TI89, =$400000 for TI92+
 addi.l #$20000,d0 ;now d0 points somewhere in TIOS
 move.l d0,12(a7) ;fool the routine to believe that it is
 move.l d0,16(a7) ;called somewhere from TIOS
 trap #$C ;this trap only enters supervisor mode
 move.w #$2700,sr ;disable ints
 move.l #$F,d3 ;prepare for calling trap #$B indirectly
 move.l #$3e000,a3 ;pass $3e000 to the trap #$B
 move.l a0,-(a7) ;push the new PC on the stack
 clr.w -(a7) ;push 0 on the stack (new SR)
 move.l $AC,a0 ;the vector for trap #$B
 jmp (a0) ;enter into trap #$B indirectly
finishtrap4: ;This will be in user mode!
 lea.l 16(a7),a7 ;free the stack space of reenter_ghost_space
 move.l #finishtrap4-reenter_ghost_space,(a7) ;restore the unprotected zone
 pea.l 4(a7)
 pea.l $5fc0
 move.l $c8,a0
 move.l memcpy*4(a0),a0 
 jsr (a0) ;call memcpy without destroying a4 
 lea.l finishtrap4-reenter_ghost_space+12(a7),a7 ;clean the stack up
 move.l (a7)+,$400ac ;restore the new trap #$B
 move.l (a7)+,d3 ;restore d3
 move.l (a7)+,a3 ;restore a3
 trap #$c ;reenter supervisor mode
 move.l usp,a0
 move.l 2(a0),-(a7)  ;push return address on the supervisor stack
 move.w (a0),-(a7) ;push return SR on the supervisor stack
 addq.l #6,a0
 move.l a0,usp ;pop them from the user stack
 rte ;return
endtrap4:

;error messages:

trap4DL: dc.b 'Please temporarily uninstall any memory resident keyboard '
         dc.b 'auto-repeat accelerators or password programs before '
         dc.b 'installing or updating h220xTSR.',0
trap4DLT: dc.b 'ERROR: Trap 4 already hooked',0
nomemDL: dc.b 'Not enough memory to install h220xTSR.',0
nomemDLT: dc.b 'ERROR: Not enough memory',0

 END
